import numpy as np

import nengo
from nengo.dists import Uniform
from nengo.params import Default, IntParam, NumberParam
from nengo_spa.network import Network
from nengo_spa.modules.scalar import Scalar
from nengo_spa.modules.state import State
from nengo.synapses import Lowpass, SynapseParam


class Thalamus(Network):
    """Inhibits non-selected actions.

    The thalamus is intended to work in tandem with a basal ganglia network.
    It converts basal ganglia output into a signal with (approximately) 1 for
    the selected action and 0 elsewhere.

    In order to suppress low responses and strengthen high responses,
    a constant bias is added to each dimension (i.e., action), and dimensions
    mutually inhibit each other. Additionally, the ensemble representing
    each dimension is created with positive encoders and can be assigned
    positive x-intercepts to threshold low responses.

    Parameters
    ----------
    neurons_action : int, optional (Default: 50)
        Number of neurons per action to represent the selection.
    threshold_action : float, optional (Default: 0.2)
        Minimum value for action representation.
    mutual_inhibit : float, optional (Default: 1.0)
        Strength of inhibition between actions.
    route_inhibit : float, optional (Default: 3.0)
        Strength of inhibition for unchosen actions.
    synapse_inhibit : float, optional (Default: 0.008)
        Synaptic filter to apply for inhibition between actions.
    synapse_bg : float, optional (Default: 0.008)
        Synaptic filter for connection between basal ganglia and thalamus.
    synapse_direct : float, optional (Default: 0.01)
        Synaptic filter for direct outputs.
    neurons_channel_dim : int, optional (Default: 50)
        Number of neurons per routing channel dimension.
    synapse_channel : float, optional (Default: 0.01)
        Synaptic filter for channel inputs and outputs.
    neurons_gate : int, optional (Default: 40)
        Number of neurons per gate.
    threshold_gate : float, optional (Default: 0.3)
        Minimum value for gating neurons.
    synapse_to-gate : float, optional (Default: 0.002)
        Synaptic filter for controlling a gate.

    kwargs
        Passed through to ``spa.Network``.

    Attributes
    ----------
    actions : EnsembleArray
        Each ensemble represents one dimension (action).
    bias : Node
        The constant bias injected in each ``actions`` ensemble.
    input : Node
        Input to the ``actions`` ensembles.
    output : Node
        Output from the ``actions`` ensembles.
    """

    neurons_action = IntParam('neurons_action', default=50)
    threshold_action = NumberParam('threshold_action', default=0.2)
    mutual_inhibit = NumberParam('mutual_inhibit', default=1.)
    route_inhibit = NumberParam('route_inhibit', default=3.)
    synapse_inhibit = SynapseParam('synapse_inhibit', default=Lowpass(0.008))
    synapse_bg = SynapseParam('synapse_bg', default=Lowpass(0.008))
    neurons_channel_dim = IntParam('neurons_channel_dim', default=50)
    synapse_channel = SynapseParam('synapse_channel', default=Lowpass(0.01))
    neurons_gate = IntParam('neurons_gate', default=40)
    threshold_gate = NumberParam('threshold_gate', default=0.3)
    synapse_to_gate = SynapseParam('synapse_to_gate', default=Lowpass(0.002))

    def __init__(self, action_count, neurons_action=Default,
                 threshold_action=Default, mutual_inhibit=Default,
                 route_inhibit=Default, synapse_inhibit=Default,
                 synapse_bg=Default, neurons_channel_dim=Default,
                 synapse_channel=Default, neurons_gate=Default,
                 threshold_gate=Default, synapse_to_gate=Default, **kwargs):
        if 'label' not in kwargs:
            kwargs['label'] = "Thalamus"
        super(Thalamus, self).__init__(**kwargs)

        self.action_count = action_count
        self.neurons_action = neurons_action
        self.mutual_inhibit = mutual_inhibit
        self.route_inhibit = route_inhibit
        self.synapse_inhibit = synapse_inhibit
        self.threshold_action = threshold_action
        self.neurons_channel_dim = neurons_channel_dim
        self.synapse_channel = synapse_channel
        self.neurons_gate = neurons_gate
        self.threshold_gate = threshold_gate
        self.synapse_to_gate = synapse_to_gate
        self.synapse_bg = synapse_bg

        self.gates = {}     # gating ensembles per action (created as needed)
        self.channels = []  # channels to pass data between networks

        self.gate_in_connections = {}
        self.gate_out_connections = {}
        self.channel_out_connections = []
        self.fixed_connections = {}
        self.bg_connection = None

        with self:
            self.actions = nengo.networks.EnsembleArray(
                self.neurons_action, self.action_count,
                intercepts=nengo.dists.Uniform(self.threshold_action, 1),
                encoders=nengo.dists.Choice([[1.0]]), label="actions")
            nengo.Connection(
                self.actions.output, self.actions.input,
                transform=(
                    np.eye(self.action_count) - 1) * self.mutual_inhibit)
            self.bias = nengo.Node([1], label="thalamus bias")
            nengo.Connection(self.bias, self.actions.input,
                             transform=np.ones((self.action_count, 1)))

        self.input = self.actions.input
        self.output = self.actions.output

    def construct_gate(self, index, net, label=None):
        """Construct a gate ensemble.

        The gate neurons have no activity when the action is selected, but are
        active when the action is not selected. This makes the gate useful for
        inhibiting ensembles that should only be active when this action is
        active.

        Parameters
        ----------
        index : int
            Index to identify the gate.
        net : :class:`nengo.Network`, optional
            Network to which to add the channel. Defaults to ``self.spa``.
        label : str, optional
            Label for the gate.

        Returns
        -------
        :class:`nengo.Ensemble`
            The constructed gate.
        """
        if label is None:
            label = 'gate[%d]' % index
        with net:
            intercepts = Uniform(self.threshold_gate, 1)
            self.gates[index] = gate = nengo.Ensemble(
                self.neurons_gate, dimensions=1, intercepts=intercepts,
                label=label, encoders=[[1]] * self.neurons_gate)
            if not hasattr(net, 'bias'):
                net.bias = nengo.Node([1], label="bias")
            nengo.Connection(net.bias, gate, synapse=None)

        self.gate_in_connections[index] = nengo.Connection(
            self.actions.ensembles[index], self.gates[index],
            synapse=self.synapse_to_gate, transform=-1)

        return self.gates[index]

    def construct_channel(
            self, target_net, target_input, net, label=None):
        """Construct a channel.

        Channels are an additional neural population in-between a source
        population and a target population. This allows inhibiting the channel
        without affecting the source and thus is useful in routing information.

        Parameters
        ----------
        target_net : :class:`spa.Network`
            The network that the channel will project to.
        target_vocab : :class:`spa.Vocabulary`
            The vocabulary used by the target population..
        net : :class:`nengo.Network`, optional
            Network to which to add the channel. Defaults to ``self.spa``.
        label : str, optional
            Label for the channel.

        Returns
        -------
        :class:`nengo.networks.EnsembleArray`
            The constructed channel.
        """
        if label is None:
            if target_net.label is not None:
                label = 'channel to ' + target_net.label
            else:
                label = 'channel'
        with net:
            if target_input[1] is None:
                channel = Scalar(label=label)
            else:
                vocab = target_input[1]
                channel = State(vocab=vocab, label=label)

        self.channels.append(channel)
        self.channel_out_connections.append(nengo.Connection(
            channel.output, target_input[0], synapse=self.synapse_channel))
        return channel

    def connect_bg(self, bg):
        """Connect a basal ganglia network to this thalamus."""
        self.bg_connection = nengo.Connection(
            bg.output, self.input, synapse=self.synapse_bg)

    def connect_gate(self, index, channel):
        """Connect a gate to a channel for information routing.

        Parameters
        ----------
        index : int
            Index of the gate to connect.
        channel : :class:`nengo.networks.EnsembleArray`
            Channel to inhibit with the gate.
        """
        if isinstance(channel, Scalar):
            target = channel.scalar.neurons
        elif isinstance(channel, State):
            target = channel.state_ensembles.add_neuron_input()
        else:
            raise NotImplementedError()

        inhibit = ([[-self.route_inhibit]] * (target.size_in))
        self.gate_out_connections[index] = nengo.Connection(
            self.gates[index], target, transform=inhibit,
            synapse=self.synapse_inhibit)

    def connect_fixed(self, index, target, transform):
        """Create connection to route fixed value.

        Parameters
        ----------
        index : int
            Index of the action to connect.
        target : :class:`nengo.base.NengoObject`
            Target of the connection.
        transform : array-like
            Transform to apply to apply to the connection.
        """
        self.fixed_connections[index] = self.connect(
            self.actions.ensembles[index], target, transform)

    def connect(self, source, target, transform):
        """Create connection.

        The connection will use the thalamus' `synapse_channel`.
        """
        return nengo.Connection(
            source, target, transform=transform, synapse=self.synapse_channel)



import pytest

import nengo_spa as spa
from nengo_spa.exceptions import SpaTypeError
from nengo_spa.actions import Parser
from nengo_spa.ast import (
    Action, ApproxInverse, DotProduct, Effect, Effects, Negative, Module,
    Product, Sink, Sum, Symbol, TAction, TEffect, TEffects, TScalar,
    TVocabulary)


def test_scalar():
    ast = Parser().parse_expr('1')
    assert ast == 1


def test_symbol():
    ast = Parser().parse_expr('A')
    assert ast == Symbol('A')
    assert str(ast) == 'A'

    vocab_type = TVocabulary(spa.Vocabulary(16, strict=False))
    ast.infer_types(None, vocab_type)
    assert ast.type == vocab_type

    ast = Parser().parse_expr('A')
    with pytest.raises(SpaTypeError):
        ast.infer_types(None, TScalar)


def test_spa_network():
    d = 16
    with spa.Network() as model:
        model.state = spa.State(d)

    ast = Parser().parse_expr('state')
    assert ast == Module('state')
    assert str(ast) == 'state'
    ast.infer_types(model, None)
    assert ast.type.vocab == model.state.vocabs[d]

    with spa.Network() as model:
        with spa.Network() as model.network:
            model.network.state = spa.State(d)

    ast = Parser().parse_expr('network.state')
    assert ast == Module('network.state')
    assert str(ast) == 'network.state'
    ast.infer_types(model, None)
    assert ast.type.vocab == model.network.state.vocabs[d]


def test_scalar_multiplication():
    ast = Parser().parse_expr('2 * A')
    assert ast == Product(2, Symbol('A'))
    assert str(ast) == '2 * A'

    vocab_type = TVocabulary(spa.Vocabulary(16, strict=False))
    ast.infer_types(None, vocab_type)
    assert ast.type == vocab_type

    ast = Parser().parse_expr('A * 2')
    assert ast == Product(Symbol('A'), 2)
    assert str(ast) == 'A * 2'

    vocab_type = TVocabulary(spa.Vocabulary(16, strict=False))
    ast.infer_types(None, vocab_type)
    assert ast.type == vocab_type

    d = 16
    with spa.Network() as model:
        model.state = spa.State(d)

    ast = Parser().parse_expr('2 * state')
    assert ast == Product(2, Module('state'))
    assert str(ast) == '2 * state'

    ast.infer_types(model, None)
    assert ast.type.vocab == model.state.vocabs[d]

    ast = Parser().parse_expr('state * 2')
    assert ast == Product(Module('state'), 2)
    assert str(ast) == 'state * 2'

    ast.infer_types(model, None)
    assert ast.type.vocab == model.state.vocabs[d]


@pytest.mark.parametrize('symbol, klass', [('+', Sum), ('*', Product)])
def test_binary_operations(symbol, klass):
    ast = Parser().parse_expr('A {} B'.format(symbol))
    assert ast == klass(Symbol('A'), Symbol('B'))
    assert str(ast) == 'A {} B'.format(symbol)

    vocab_type = TVocabulary(spa.Vocabulary(16, strict=False))
    ast.infer_types(None, vocab_type)
    assert ast.type == vocab_type

    d = 16
    with spa.Network() as model:
        model.state = spa.State(d)
        model.state2 = spa.State(d)

    ast = Parser().parse_expr('state {} B'.format(symbol))
    assert ast == klass(Module('state'), Symbol('B'))
    assert str(ast) == 'state {} B'.format(symbol)

    ast.infer_types(model, TVocabulary(model.state.vocabs[d]))
    assert ast.type.vocab == model.state.vocabs[d]

    ast = Parser().parse_expr('A {} state'.format(symbol))
    assert ast == klass(Symbol('A'), Module('state'))
    assert str(ast) == 'A {} state'.format(symbol)

    ast.infer_types(model, TVocabulary(model.state.vocabs[d]))
    assert ast.type.vocab == model.state.vocabs[d]

    ast = Parser().parse_expr('state {} state2'.format(symbol))
    assert ast == klass(Module('state'), Module('state2'))
    assert str(ast) == 'state {} state2'.format(symbol)

    ast.infer_types(model, None)
    assert ast.type.vocab == model.state.vocabs[d]


@pytest.mark.parametrize('symbol, klass', [
    ('~', ApproxInverse), ('-', Negative)])
def test_unary(symbol, klass):
    ast = Parser().parse_expr(symbol + 'A')
    assert ast == klass(Symbol('A'))
    assert str(ast) == symbol + 'A'

    vocab_type = TVocabulary(spa.Vocabulary(16, strict=False))
    ast.infer_types(None, vocab_type)
    assert ast.type == vocab_type

    d = 16
    with spa.Network() as model:
        model.state = spa.State(d)

    ast = Parser().parse_expr(symbol + 'state')
    assert ast == klass(Module('state'))
    assert str(ast) == symbol + 'state'

    ast.infer_types(model, None)
    assert ast.type.vocab == model.state.vocabs[d]


def test_dot_product():
    d = 16
    with spa.Network() as model:
        model.state = spa.State(d)

    ast = Parser().parse_expr('dot(A, state)')
    assert ast == DotProduct(Symbol('A'), Module('state'))
    assert str(ast) == 'dot(A, state)'

    ast.infer_types(model, TVocabulary(model.state.vocabs[d]))
    assert ast.type == TScalar

    ast = Parser().parse_expr('2 * dot(A, state) + 1')
    assert ast == Sum(Product(2, DotProduct(Symbol('A'), Module('state'))), 1)
    assert str(ast) == '2 * dot(A, state) + 1'

    ast.infer_types(model, TVocabulary(model.state.vocabs[d]))
    assert ast.type == TScalar


def test_effect():
    d = 16
    with spa.Network() as model:
        model.state = spa.State(d)

    ast = Parser().parse_effect('state = A')
    assert ast == Effect(Sink('state'), Symbol('A'))
    assert str(ast) == 'state = A'
    assert ast.type == TEffect


def test_effects():
    # Check that multiple lvalue=rvalue parsing is working with commas
    ast = Parser().parse_effects('x=a,y=dot(a,b),z=b')
    assert ast == Effects(
        Effect(Sink('x'), Module('a')),
        Effect(Sink('y'), DotProduct(Module('a'), Module('b'))),
        Effect(Sink('z'), Module('b')))
    assert str(ast) == 'x = a, y = dot(a, b), z = b'
    assert ast.type == TEffects

    ast = Parser().parse_effects('  foo = dot(a, b)  , bar = b')
    assert ast == Effects(
        Effect(Sink('foo'), DotProduct(Module('a'), Module('b'))),
        Effect(Sink('bar'), Module('b')))
    assert str(ast) == 'foo = dot(a, b), bar = b'


def test_action():
    d = 16
    with spa.Network() as model:
        model.state = spa.State(d)

    ast = Parser().parse_action('dot(state, A) --> state = B')
    assert ast == Action(
        DotProduct(Module('state'), Symbol('A')),
        Effects(Effect(Sink('state'), Symbol('B'), channeled=True)))
    assert str(ast) == 'dot(state, A) --> state = B'
    assert ast.type == TAction


def test_complex_epressions():
    d = 16
    with spa.Network() as model:
        model.state = spa.State(d)

    ast = Parser().parse_expr('~(A - B * state)')
    assert ast == ApproxInverse(Sum(
        Symbol('A'), Negative(Product(Symbol('B'), Module('state')))))
    assert str(ast) == '~(A + -(B * state))'

    ast.infer_types(model, TVocabulary(model.state.vocabs[d]))
    assert ast.type.vocab == model.state.vocabs[d]

    ast = Parser().parse_expr('0.5*(2*dot(a, A)-dot(b,B))-2')
    assert str(ast) == '0.5 * (2 * dot(a, A) + -dot(b, B)) + -2'

    ast = Parser().parse_expr('dot(x, -1) + 1')
    assert str(ast) == 'dot(x, -1) + 1'

    ast = Parser().parse_expr('2*dot(a, 1) - dot(b, -1) + dot(a, b)')
    assert str(ast) == '2 * dot(a, 1) + -dot(b, -1) + dot(a, b)'

    ast = Parser().parse_expr('a*b - 1 + 2*b')
    assert str(ast) == 'a * b + -1 + 2 * b'


def test_zero_vector():
    d = 16
    with spa.Network() as model:
        model.state = spa.State(d)

    ast = Parser().parse_effect('state = 0')
    ast.infer_types(model, None)
    assert ast.source.type.vocab == model.state.vocabs[d]


def test_vocab_transform_in_multiplication():
    d = 16
    with spa.Network() as model:
        model.state = spa.State(d)

    ast = Parser().parse_expr('2 * translate(state)')
    assert str(ast) == '2 * translate(state)'

    ast = Parser().parse_expr('translate(state) * 2')
    assert str(ast) == 'translate(state) * 2'

    ast = Parser().parse_expr('2 * reinterpret(state)')
    assert str(ast) == '2 * reinterpret(state)'

    ast = Parser().parse_expr('reinterpret(state) * 2')
    assert str(ast) == 'reinterpret(state) * 2'



import nengo


def make_parse_func(func, vocab):
    """Create a function that calls func and parses the output in vocab."""

    def parse_func(t):
        return vocab.parse(func(t)).v

    return parse_func


class _HierachicalInputProxy(object):
    def __init__(self, parent, name):
        self.__dict__['parent'] = parent
        self.__dict__['name'] = name

    def __getattr__(self, name):
        return _HierachicalInputProxy(self.parent, self.name + '.' + name)

    def __setattr__(self, name, value):
        setattr(self.parent, self.name + '.' + name, value)


class Input(nengo.Network):
    """A SPA network for providing external inputs to other networks.

    The parameters passed to this network indicate the network input name
    and the function to execute to generate inputs to that network.
    The functions should always return strings, which will then be parsed
    by the relevant vocabulary. For example::

        def input1(t):
            if t < 0.1:
                return 'A'
            else:
                return '0'

        spa_net.input = spa.Input(vision=input1, task='X')

    will create two inputs:

    1. an input to the ``vision`` network, which for the first 0.1 seconds
       is the value associated with the ``'A'`` semantic pointer and then
       a vector of all zeros, and
    2. an input to the ``task`` network which is always the value associated
       with the ``'X'`` semantic pointer.

    Parameters
    ----------
    network : spa.Network, optional (Default: the current SPA network)
        Network that this instance provides input for.
    kwargs
        Keyword arguments passed through to ``nengo.Network``.
    """

    def __init__(self, network=None, **kwargs):
        super(Input, self).__init__(**kwargs)
        self.input_nodes = {}

        if network is None:
            from nengo_spa.network import get_current_spa_network
            network = get_current_spa_network()
        self.network = network

        self._initialized = True

    def __connect(self, name, expr):
        target, vocab = self.network.get_network_input(name)
        if callable(expr):
            val = make_parse_func(expr, vocab)
        else:
            val = vocab.parse(expr).v

        with self:
            node = nengo.Node(val, label='input_%s' % name)
        self.input_nodes[name] = node

        with self.network:
            nengo.Connection(node, target, synapse=None)

    def __setattr__(self, name, value):
        if not getattr(self, '_initialized') or name in self.__dict__:
            super(Input, self).__setattr__(name, value)
        else:
            self.__connect(name, value)

    def __getattr__(self, name):
        if name == '_initialized':
            return self.__dict__.get('_initialized', False)
        return _HierachicalInputProxy(self, name)



from nengo.params import BoolParam, Default, IntParam
import nengo_spa
from nengo_spa.network import Network
from nengo_spa.vocab import VocabularyOrDimParam


class Bind(Network):
    """A network for binding together two inputs.

    Binding is done with circular convolution. For more details on how
    this is computed, see the underlying `~.network.CircularConvolution`
    network.

    Parameters
    ----------
    vocab : Vocabulary or int
        The vocabulary to use to interpret the vector. If an integer is given,
        the default vocabulary of that dimensionality will be used.
    neurons_per_dimension : int, optional (Default: 200)
        Number of neurons to use in each product computation.
    invert_a, invert_b : bool, optional (Default: False, False)
        Whether to reverse the order of elements in either
        the first input (``invert_a``) or the second input (``invert_b``).
        Flipping the second input will make the network perform circular
        correlation instead of circular convolution.
    kwargs
        Keyword arguments passed through to ``spa.Network``.
    """

    vocab = VocabularyOrDimParam('vocab', default=None, readonly=True)
    neurons_per_dimension = IntParam(
        'neurons_per_dimension', default=200, low=1, readonly=True)
    invert_a = BoolParam('invert_a', default=False, readonly=True)
    invert_b = BoolParam('invert_b', default=False, readonly=True)

    def __init__(self, vocab=Default, neurons_per_dimension=Default,
                 invert_a=Default, invert_b=Default, **kwargs):
        super(Bind, self).__init__(**kwargs)

        self.vocab = vocab
        self.neurons_per_dimension = neurons_per_dimension
        self.invert_a = invert_a
        self.invert_b = invert_b

        with self:
            self.cc = nengo_spa.networks.CircularConvolution(
                self.neurons_per_dimension, self.vocab.dimensions,
                self.invert_a, self.invert_b)

        self.input_a = self.cc.input_a
        self.input_b = self.cc.input_b
        self.output = self.cc.output

        self.inputs = dict(
            input_a=(self.input_a, self.vocab),
            input_b=(self.input_b, self.vocab))
        self.outputs = dict(default=(self.output, self.vocab))



import numpy as np

import nengo
import nengo_spa as spa


def test_basic():
    with spa.Network() as model:
        model.state = spa.State(vocab=16)

    input = model.get_network_input('state')
    output = model.get_network_output('state')
    assert input[0] is model.state.input
    assert output[0] is model.state.output
    assert input[1] is output[1]
    assert input[1].dimensions == 16


def test_neurons():
    with spa.Network() as model:
        model.state = spa.State(
            vocab=16, neurons_per_dimension=2, represent_identity=False)

    assert len(model.state.state_ensembles.ensembles) == 1
    assert model.state.state_ensembles.ensembles[0].n_neurons == 16 * 2

    with spa.Network() as model:
        model.state = spa.State(
            vocab=16, subdimensions=1, neurons_per_dimension=2,
            represent_identity=False)

    assert len(model.state.state_ensembles.ensembles) == 16
    assert model.state.state_ensembles.ensembles[0].n_neurons == 2


def test_no_feedback_run(Simulator, plt, seed):
    with spa.Network(seed=seed) as model:
        model.state = spa.State(vocab=32, feedback=0.0)

        def state_input(t):
            if 0 <= t < 0.3:
                return 'A'
            elif 0.2 <= t < 0.6:
                return 'B'
            else:
                return '0'
        model.state_input = spa.Input()
        model.state_input.state = state_input

    state, vocab = model.get_network_output('state')

    with model:
        p = nengo.Probe(state, 'output', synapse=0.05)

    with Simulator(model) as sim:
        sim.run(0.8)

    data = np.dot(sim.data[p], vocab.vectors.T)
    plt.plot(sim.trange(), data)
    assert data[299, 0] > 0.9
    assert data[299, 1] < 0.2
    assert data[599, 0] < 0.2
    assert data[599, 1] > 0.9
    assert data[799, 0] < 0.2
    assert data[799, 1] < 0.2


def test_memory_run(Simulator, seed, plt):
    with spa.Network(seed=seed) as model:
        model.memory = spa.State(vocab=32, feedback=1.0,
                                 feedback_synapse=0.01)

        def state_input(t):
            if 0 <= t < 0.05:
                return 'A'
            else:
                return '0'

        model.state_input = spa.Input()
        model.state_input.memory = state_input

    memory, vocab = model.get_network_output('memory')

    with model:
        p = nengo.Probe(memory, 'output', synapse=0.03)

    with Simulator(model) as sim:
        sim.run(0.5)
    t = sim.trange()

    similarity = np.dot(sim.data[p], vocab.vectors.T)
    plt.plot(t, similarity)
    plt.ylabel("Similarity to 'A'")
    plt.xlabel("Time (s)")

    # value should peak above 1.0, then decay down to near 1.0
    assert np.mean(similarity[(t > 0.05) & (t < 0.1)]) > 1.0
    assert np.mean(similarity[(t > 0.2) & (t < 0.3)]) > 0.7
    assert np.mean(similarity[t > 0.49]) > 0.5


def test_memory_run_decay(Simulator, plt, seed):
    with spa.Network(seed=seed) as model:
        model.memory = spa.State(vocab=32, feedback=(1.0 - 0.01/0.05),
                                 feedback_synapse=0.01)

        def state_input(t):
            if 0 <= t < 0.05:
                return 'A'
            else:
                return '0'

        model.state_input = spa.Input()
        model.state_input.memory = state_input

    memory, vocab = model.get_network_output('memory')

    with model:
        p = nengo.Probe(memory, 'output', synapse=0.03)

    with Simulator(model) as sim:
        sim.run(0.3)
    data = np.dot(sim.data[p], vocab.vectors.T)

    t = sim.trange()
    plt.plot(t, data)

    assert data[t == 0.05, 0] > 1.0
    assert data[t == 0.299, 0] < 0.4


def test_represent_identity(Simulator, seed):
    d = 32
    with spa.Network(seed=seed) as model:
        model.memory = spa.State(d, represent_identity=True)
        model.input = spa.Input()
        model.input.memory = '1'
        p = nengo.Probe(model.memory.output, synapse=0.03)

    with Simulator(model) as sim:
        sim.run(0.3)

    assert np.all(sim.data[p][sim.trange() > 0.2][:, 0] > 0.9)
    assert np.all(np.abs(sim.data[p][sim.trange() > 0.2][:, 1:]) < 0.1)



#!/usr/bin/env python
import imp
import io
import os
import sys

try:
    from setuptools import find_packages, setup
except ImportError:
    raise ImportError(
        "'setuptools' is required but not installed. To install it, "
        "follow the instructions at "
        "https://pip.pypa.io/en/stable/installing/#installing-with-get-pip-py")


def read(*filenames, **kwargs):
    encoding = kwargs.get('encoding', 'utf-8')
    sep = kwargs.get('sep', '\n')
    buf = []
    for filename in filenames:
        with io.open(filename, encoding=encoding) as f:
            buf.append(f.read())
    return sep.join(buf)


root = os.path.dirname(os.path.realpath(__file__))
version_module = imp.load_source(
    'version', os.path.join(root, 'nengo_spa', 'version.py'))
testing = 'test' in sys.argv or 'pytest' in sys.argv

setup(
    name="nengo_spa",
    version=version_module.version,
    author="Applied Brain Research",
    author_email="info@appliedbrainresearch.com",
    packages=find_packages(),
    scripts=[],
    license="Free for non-commercial use",
    description="An implementation of the Semantic Pointer Architecture for "
                "Nengo",
    long_description=read('README.rst', 'CHANGES.rst'),
    zip_safe=True,
    include_package_data=True,
    setup_requires=['pytest-runner'] if testing else [] + [
        'numpy>=1.7',
    ],
    install_requires=[
        'nengo>=2.4',
        'numpy>=1.7',
    ],
    extras_require={
        'prob_cleanup': ['scipy'],
    },
    tests_require=[
        'jupyter',
        'matplotlib>=1.4',
        'pytest>=2.3',
    ],
    entry_points={
    },
    classifiers=[  # https://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'License :: Free for non-commercial use',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
    ],
)



"""SPA modules.

SPA modules derive from `nengo_spa.Network` and thus provide information about
inputs and outputs that can be used in action rules and the associated
vocabularies. Note that a module might have no inputs and outputs that can be
used in action rules (like `BasalGanglia` and `Thalamus`), but only inputs and
outputs that must be manually connected to. Many SPA modules are networks that
might be automatically created by building action rules. Because of this, it is
possible to set module parameters with `nengo.Config` objects to allow to
easily change parameters of networks created in this way.

Note that SPA modules can be used as standalone networks without using
nengo_spa features.
"""

from .assoc_mem import (
    AssociativeMemory,
    IAAssocMem,
    ThresholdingAssocMem,
    WTAAssocMem)
from .basalganglia import BasalGanglia
from .bind import Bind
from .compare import Compare
from .input import Input
from .product import Product
from .scalar import Scalar
from .state import State
from .thalamus import Thalamus



"""Functions to examine data and vectors produced by SPA."""

from itertools import combinations

import numpy as np

import nengo.utils.numpy as npext
from nengo.exceptions import ValidationError
from nengo_spa.pointer import SemanticPointer
from nengo.utils.compat import is_iterable


def similarity(data, vocab, normalize=False):
    """Return the similarity between some data and the vocabulary.

    Computes the dot products between all data vectors and each
    vocabulary vector. If ``normalize=True``, normalizes all vectors
    to compute the cosine similarity.

    Parameters
    ----------
    data: array_like
        The data used for comparison.
    vocab: Vocabulary or array_like
        Vocabulary (or list of vectors) to use to calculate
        the similarity values.
    normalize : bool, optional (Default: False)
        Whether to normalize all vectors, to compute the cosine similarity.
    """
    from nengo_spa.vocab import Vocabulary

    if isinstance(data, SemanticPointer):
        data = data.v

    if isinstance(vocab, Vocabulary):
        vectors = vocab.vectors
    elif is_iterable(vocab):
        if isinstance(next(iter(vocab)), SemanticPointer):
            vocab = [p.v for p in vocab]
        vectors = np.array(vocab, copy=False, ndmin=2)
    else:
        raise ValidationError("%r object is not a valid vocabulary"
                              % (type(vocab).__name__), attr='vocab')

    dots = np.dot(vectors, data.T)

    if normalize:
        # Zero-norm vectors should return zero, so avoid divide-by-zero error
        eps = np.nextafter(0, 1)  # smallest float above zero
        dnorm = np.maximum(npext.norm(data.T, axis=0, keepdims=True), eps)
        vnorm = np.maximum(npext.norm(vectors, axis=1, keepdims=True), eps)

        if len(dots.shape) == 1:
            vnorm = np.squeeze(vnorm)

        dots /= dnorm
        dots /= vnorm

    return dots.T


def pairs(vocab):
    return set(x + '*' + y for x, y in combinations(vocab.keys(), 2))


def text(v, vocab, minimum_count=1, maximum_count=None,
         threshold=0.1, join=';', terms=None, normalize=False):
    """Return a human-readable text version of the provided vector.

    This is meant to give a quick text version of a vector for display
    purposes. To do this, compute the dot product between the vector
    and all the terms in the vocabulary. The top few vectors are
    chosen for inclusion in the text. It will try to only return
    terms with a match above the threshold, but will always return
    at least minimum_count and at most maximum_count terms. Terms
    are sorted from most to least similar.

    Parameters
    ----------
    v : SemanticPointer or ndarray
        The vector to convert into text.
    minimum_count : int, optional (Default: 1)
        Always return at least this many terms in the text.
    maximum_count : int, optional (Default: None)
        Never return more than this many terms in the text.
        If None, all terms will be returned.
    threshold : float, optional (Default: 0.1)
        How small a similarity for a term to be ignored.
    join : str, optional (Default: ';')
        The text separator to use between terms.
    terms : list, optional (Default: None)
        Only consider terms in this list of strings.
    normalize : bool, optional (Default: False)
        Whether to normalize the vector before computing similarity.
    """
    if not isinstance(v, SemanticPointer):
        v = SemanticPointer(v)
    if normalize:
        v = v.normalized()

    if terms is None:
        terms = vocab.keys()
        vectors = vocab.vectors
    else:
        vectors = vocab.parse_n(*terms)

    matches = list(zip(similarity(v, vectors), terms))
    matches.sort()
    matches.reverse()

    r = []
    for m in matches:
        if minimum_count is not None and len(r) < minimum_count:
            r.append(m)
        elif maximum_count is not None and len(r) == maximum_count:
            break
        elif threshold is None or m[0] > threshold:
            r.append(m)
        else:
            break

    return join.join(['%0.2f%s' % (sim, key) for (sim, key) in r])



import numpy as np

import nengo
from nengo.exceptions import ValidationError
from nengo.networks.product import Product
from nengo.utils.compat import range
from nengo.utils.magic import memoize


def circconv(a, b, invert_a=False, invert_b=False, axis=-1):
    """A reference Numpy implementation of circular convolution"""
    A = np.fft.fft(a, axis=axis)
    B = np.fft.fft(b, axis=axis)
    if invert_a:
        A = A.conj()
    if invert_b:
        B = B.conj()
    return np.fft.ifft(A * B, axis=axis).real


@memoize
def transform_in(dims, align, invert):
    """Create a transform to map the input into the Fourier domain.

    See CircularConvolution docstring for more details.

    Parameters
    ----------
    dims : int
        Input dimensions.
    align : 'A' or 'B'
        How to align the real and imaginary components; the alignment
        depends on whether we're doing transformA or transformB.
    invert : bool
        Whether to reverse the order of elements.
    """
    if align not in ('A', 'B'):
        raise ValidationError("'align' must be either 'A' or 'B'", 'align')

    dims2 = 4 * (dims // 2 + 1)
    tr = np.zeros((dims2, dims))
    dft = dft_half(dims)

    for i in range(dims2):
        row = dft[i // 4] if not invert else dft[i // 4].conj()
        if align == 'A':
            tr[i] = row.real if i % 2 == 0 else row.imag
        else:  # align == 'B'
            tr[i] = row.real if i % 4 == 0 or i % 4 == 3 else row.imag

    remove_imag_rows(tr)
    tr /= np.sqrt(dims)
    return tr.reshape((-1, dims))


def transform_out(dims):
    dims2 = (dims // 2 + 1)
    tr = np.zeros((dims2, 4, dims))
    idft = dft_half(dims).conj()

    for i in range(dims2):
        row = idft[i] if i == 0 or 2*i == dims else 2*idft[i]
        tr[i, 0] = row.real
        tr[i, 1] = -row.real
        tr[i, 2] = -row.imag
        tr[i, 3] = -row.imag

    tr = tr.reshape(4*dims2, dims)
    remove_imag_rows(tr)
    return tr.T


def remove_imag_rows(tr):
    """Throw away imaginary row we don't need (since they're zero)"""
    i = np.arange(tr.shape[0])
    if tr.shape[1] % 2 == 0:
        tr = tr[(i == 0) | (i > 3) & (i < len(i) - 3)]
    else:
        tr = tr[(i == 0) | (i > 3)]


@memoize
def dft_half(n):
    x = np.arange(n)
    w = np.arange(n // 2 + 1)
    return np.exp((-2.j * np.pi / n) * (w[:, None] * x[None, :]))


def CircularConvolution(n_neurons, dimensions, invert_a=False, invert_b=False,
                        input_magnitude=1.0, net=None):
    r"""Compute the circular convolution of two vectors.

    The circular convolution :math:`c` of vectors :math:`a` and :math:`b`
    is given by

    .. math:: c[i] = \sum_j a[j] b[i - j]

    where negative indices on :math:`b` wrap around to the end of the vector.

    This computation can also be done in the Fourier domain,

    .. math:: c = DFT^{-1} ( DFT(a) DFT(b) )

    where :math:`DFT` is the Discrete Fourier Transform operator, and
    :math:`DFT^{-1}` is its inverse. This network uses this method.

    Parameters
    ----------
    n_neurons : int
        Number of neurons to use in each product computation
    dimensions : int
        The number of dimensions of the input and output vectors.

    invert_a, invert_b : bool, optional (Default: False, False)
        Whether to reverse the order of elements in either
        the first input (``invert_a``) or the second input (``invert_b``).
        Flipping the second input will make the network perform circular
        correlation instead of circular convolution.
    input_magnitude : float, optional (Default: 1.0)
        The expected magnitude of the vectors to be convolved.
        This value is used to determine the radius of the ensembles
        computing the element-wise product.
    net : Network, optional (Default: None)
        A network in which the network components will be built.
        This is typically used to provide a custom set of Nengo object
        defaults through modifying ``net.config``.

    Returns
    -------
    net : Network
        The newly built product network, or the provided ``net``.

    Attributes
    ----------
    net.input_a : Node
        The first vector to be convolved.
    net.input_b : Node
        The second vector to be convolved.
    net.product : Network
        Network created with `.Product` to do the element-wise product
        of the :math:`DFT` components.
    net.output : Node
        The resulting convolved vector.

    Examples
    --------

    A basic example computing the circular convolution of two 10-dimensional
    vectors represented by ensemble arrays::

        A = EnsembleArray(50, n_ensembles=10)
        B = EnsembleArray(50, n_ensembles=10)
        C = EnsembleArray(50, n_ensembles=10)
        cconv = nengo.networks.CircularConvolution(50, dimensions=10)
        nengo.Connection(A.output, cconv.input_a)
        nengo.Connection(B.output, cconv.input_b)
        nengo.Connection(cconv.output, C.input)

    Notes
    -----

    The network maps the input vectors :math:`a` and :math:`b` of length N into
    the Fourier domain and aligns them for complex multiplication.
    Letting :math:`F = DFT(a)` and :math:`G = DFT(b)`, this is given by::

        [ F[i].real ]     [ G[i].real ]     [ w[i] ]
        [ F[i].imag ]  *  [ G[i].imag ]  =  [ x[i] ]
        [ F[i].real ]     [ G[i].imag ]     [ y[i] ]
        [ F[i].imag ]     [ G[i].real ]     [ z[i] ]

    where :math:`i` only ranges over the lower half of the spectrum, since
    the upper half of the spectrum is the flipped complex conjugate of
    the lower half, and therefore redundant. The input transforms are
    used to perform the DFT on the inputs and align them correctly for
    complex multiplication.

    The complex product :math:`H = F * G` is then

    .. math:: H[i] = (w[i] - x[i]) + (y[i] + z[i]) I

    where :math:`I = \sqrt{-1}`. We can perform this addition along with the
    inverse DFT :math:`c = DFT^{-1}(H)` in a single output transform, finding
    only the real part of :math:`c` since the imaginary part
    is analytically zero.
    """
    if net is None:
        net = nengo.Network("Circular Convolution")

    tr_a = transform_in(dimensions, 'A', invert_a)
    tr_b = transform_in(dimensions, 'B', invert_b)
    tr_out = transform_out(dimensions)

    with net:
        net.input_a = nengo.Node(size_in=dimensions, label="input_a")
        net.input_b = nengo.Node(size_in=dimensions, label="input_b")
        with nengo.Config(nengo.Ensemble) as cfg:
            cfg[nengo.Ensemble].eval_points = nengo.dists.CosineSimilarity(
                2 * dimensions + 2)
            cfg[nengo.Ensemble].intercepts = nengo.dists.CosineSimilarity(
                2 * dimensions + 2)
            net.product = Product(
                n_neurons, tr_out.shape[1],
                input_magnitude=input_magnitude / np.sqrt(2.))
        net.output = nengo.Node(size_in=dimensions, label="output")

        nengo.Connection(
            net.input_a, net.product.input_a, transform=tr_a, synapse=None)
        nengo.Connection(
            net.input_b, net.product.input_b, transform=tr_b, synapse=None)
        nengo.Connection(
            net.product.output, net.output, transform=tr_out, synapse=None)

    return net



import numpy as np
import pytest

import nengo
import nengo_spa as spa
from nengo_spa.exceptions import SpaTypeError


def test_new_action_syntax(Simulator, seed, plt, rng):
    model = spa.Network(seed=seed)
    model.config[spa.State].vocab = 3
    model.config[spa.State].subdimensions = 3
    with model:
        model.ctrl = spa.State(16, subdimensions=16, label='ctrl')

        def input_func(t):
            if t < 0.2:
                return 'A'
            elif t < 0.4:
                return 'B'
            else:
                return 'C'
        model.input = spa.Input()
        model.input.ctrl = input_func

        model.state = spa.State(label='state')
        model.buff1 = spa.State(label='buff1')
        model.buff2 = spa.State(label='buff2')
        model.buff3 = spa.State(label='buff3')

        node1 = nengo.Node([0, 1, 0])
        node2 = nengo.Node([0, 0, 1])

        nengo.Connection(node1, model.buff1.input)
        nengo.Connection(node2, model.buff2.input)

        actions = spa.Actions(
            'state = buff1',
            'dot(ctrl, A) --> buff3=buff1',
            'dot(ctrl, B) --> buff3=buff2',
            'dot(ctrl, C) --> buff3=buff1*buff2')
        actions.build()

        state_probe = nengo.Probe(model.state.output, synapse=0.03)
        buff3_probe = nengo.Probe(model.buff3.output, synapse=0.03)

    with Simulator(model) as sim:
        sim.run(0.6)

    data = sim.data[buff3_probe]

    plt.plot(sim.trange(), data)

    state_val = np.mean(sim.data[state_probe][200:], axis=0)
    assert state_val[0] < 0.2
    assert state_val[1] > 0.8
    assert state_val[2] < 0.2

    valueA = np.mean(data[150:200], axis=0)  # should be [0, 1, 0]
    valueB = np.mean(data[350:400], axis=0)  # should be [0, 0, 1]
    valueC = np.mean(data[550:600], axis=0)  # should be [1, 0, 0]

    assert valueA[0] < 0.2
    assert valueA[1] > 0.75
    assert valueA[2] < 0.2

    assert valueB[0] < 0.2
    assert valueB[1] < 0.2
    assert valueB[2] > 0.75

    assert valueC[0] > 0.75
    assert valueC[1] < 0.2
    assert valueC[2] < 0.2


def test_dot_product(Simulator, seed, plt):
    d = 16

    with spa.Network(seed=seed) as model:
        model.state_a = spa.State(d)
        model.state_b = spa.State(d)
        model.result = spa.Scalar()

        model.stimulus = spa.Input()
        model.stimulus.state_a = 'A'
        model.stimulus.state_b = lambda t: 'A' if t <= 0.3 else 'B'

        actions = spa.Actions('result = dot(state_a, state_b)')
        actions.build()

        p = nengo.Probe(model.result.output, synapse=0.03)

    with Simulator(model) as sim:
        sim.run(0.6)

    plt.plot(sim.trange(), sim.data[p])

    t = sim.trange()
    assert np.mean(sim.data[p][(t > 0.1) & (t <= 0.3)]) > 0.8
    assert np.mean(sim.data[p][t > 0.4]) < 0.1


class TestExceptions():
    @pytest.fixture
    def model(self):
        with spa.Network() as model:
            model.state_a = spa.State(16)
            model.state_b = spa.State(32)
            model.state_c = spa.State(32)
            model.scalar = spa.Scalar()
        return model

    def test_invalid_types_in_binary_operation(self, model):
        with model:
            actions = spa.Actions('state_c = state_a + state_b')
            with pytest.raises(SpaTypeError):
                actions.build()

    def test_approx_inverse_of_scalar(self, model):
        with model:
            actions = spa.Actions('state_c = ~scalar')
            with pytest.raises(SpaTypeError):
                actions.build()

    def test_dot_product_incompatiple_vocabs(self, model):
        with model:
            actions = spa.Actions('scalar = dot(state_a, state_b)')
            with pytest.raises(SpaTypeError):
                actions.build()

    def test_dot_product_first_argument_invalid(self, model):
        with model:
            actions = spa.Actions('scalar = dot(scalar, state_b)')
            with pytest.raises(SpaTypeError):
                actions.build()

    def test_dot_product_second_argument_invalid(self, model):
        with model:
            actions = spa.Actions('scalar = dot(state_a, scalar)')
            with pytest.raises(SpaTypeError):
                actions.build()

    @pytest.mark.parametrize('method', ['reinterpret', 'translate'])
    def test_cast_type_inference_not_possible(self, model, method):
        with model:
            actions = spa.Actions('scalar = dot({}(state_a), A)'.format(
                method))
            with pytest.raises(SpaTypeError):
                actions.build()

    @pytest.mark.parametrize('method', ['reinterpret', 'translate'])
    def test_cast_scalar(self, model, method):
        with model:
            actions = spa.Actions('state_a = {}(scalar)'.format(
                method))
            with pytest.raises(SpaTypeError):
                actions.build()

    def test_reinterpret_non_matching_dimensions(self, model):
        with model:
            actions = spa.Actions('state_a = reinterpret(state_b)')
            with pytest.raises(SpaTypeError):
                actions.build()


def test_access_actions():
    actions = spa.Actions('a = b', 'b = c', named='c = d')

    assert len(actions) == 3
    assert str(actions[0]) == 'a = b'
    assert str(actions[1]) == 'b = c'
    assert str(actions[2]) == 'c = d'
    assert str(actions['named']) == 'c = d'


def test_provides_access_to_constructed_objects_of_effect():
    with spa.Network() as model:
        model.config[spa.State].vocab = 16
        model.a = spa.State()
        model.b = spa.State()
        model.c = spa.State()

        actions = spa.Actions('c = a * b')
        bg, thalamus, constructed = actions.build()

        assert len(constructed[actions[0].effects[0]]) == 1
        assert isinstance(
            constructed[actions[0].effects[0]][0], nengo.Connection)
        assert len(constructed[actions[0].effects[0].source]) == 3
        n_connections = 0
        n_bind = 0
        for obj in constructed[actions[0].effects[0].source]:
            if isinstance(obj, nengo.Connection):
                n_connections += 1
            elif isinstance(obj, spa.Bind):
                n_bind += 1
            else:
                raise AssertionError("Unexpected object constructed for Bind.")
        assert n_connections == 2 and n_bind == 1



"""Functions to evaluate analytically derived equations related to the SPA."""

from nengo.dists import CosineSimilarity


def prob_cleanup(similarity, dimensions, vocab_size):
    """Estimate the chance of successful cleanup.

    This returns the chance that, out of vocab_size randomly chosen
    vectors, at least one of them will be closer to a particular
    vector than the value given by compare. To use this, compare
    your noisy vector with the ideal vector, pass that value in as
    the similarity parameter, and set ``vocab_size`` to be the number of
    competing vectors.

    Requires SciPy.
    """
    return CosineSimilarity(dimensions).cdf(similarity) ** vocab_size



import nengo
from nengo.params import Default, IntParam
from nengo_spa.network import Network


class Scalar(Network):
    """A SPA network capable of representing a single scalar.

    Parameters
    ----------
    n_neurons : int, optional (Default: 50)
        Number of neurons to represent the scalar.
    """

    n_neurons = IntParam('n_neurons', default=50, low=1, readonly=True)

    def __init__(self, n_neurons=Default, **kwargs):
        super(Scalar, self).__init__(**kwargs)

        self.n_neurons = n_neurons

        with self:
            self.scalar = nengo.Ensemble(self.n_neurons, 1)

        self.input = self.scalar
        self.output = self.scalar
        self.inputs = dict(default=(self.input, None))
        self.outputs = dict(default=(self.output, None))



import numpy as np

import nengo
from nengo.config import Config, SupportDefaultsMixin
from nengo_spa.exceptions import SpaNetworkError
from nengo_spa.modules.input import Input
from nengo_spa.vocab import VocabularyMap, VocabularyMapParam
from nengo.utils.compat import iteritems
import nengo_spa


def get_current_spa_network():
    for net in reversed(Network.context):
        if isinstance(net, Network):
            return net
    return None


class _AutoConfig(object):
    def __init__(self, cfg):
        self._cfg = cfg

    def __getattr__(self, name):
        return getattr(self._cfg, name)

    def __getitem__(self, key):
        if key not in self._cfg.params:
            self._cfg.configures(key)
        return self._cfg[key]


class Network(nengo.Network, SupportDefaultsMixin):
    """Base class for SPA networks.

    SPA networks are networks that also have a list of inputs and outputs,
    each with an associated `.Vocabulary` (or a desired dimensionality for
    the vocabulary).

    The inputs and outputs are dictionaries that map a name to an
    (object, Vocabulary) pair. The object can be a `.Node` or an `.Ensemble`.
    """

    vocabs = VocabularyMapParam('vocabs', default=None, optional=False)

    def __init__(
            self, label=None, seed=None, add_to_container=None, vocabs=None):
        super(Network, self).__init__(label, seed, add_to_container)
        self.config.configures(Network)

        if vocabs is None:
            vocabs = Config.default(Network, 'vocabs')
            if vocabs is None:
                if seed is not None:
                    rng = np.random.RandomState(seed)
                else:
                    rng = None
                vocabs = VocabularyMap(rng=rng)
        self.vocabs = vocabs
        self.config[Network].vocabs = vocabs

        self.__dict__['_parent_spa_network'] = get_current_spa_network()

        self._spa_networks = {}

        self.inputs = {}
        self.outputs = {}

        self._stimuli = None

        self._initialized = True

    @property
    def config(self):
        return _AutoConfig(self._config)

    # FIXME remove?
    @property
    def stimuli(self):
        if self._stimuli is None:
            self._stimuli = Input(self)
        return self._stimuli

    def __setstate__(self, state):
        if '_initialized' in state:
            del state['_initialized']
        super(Network, self).__setstate__(state)
        setattr(self, '_initialized', True)

    def __setattr__(self, key, value):
        """A setattr that handles SPA networks being added specially.

        This is so that we can use the variable name for the Network as
        the name that all of the SPA system will use to access that network.
        """
        if not hasattr(self, '_initialized'):
            return super(Network, self).__setattr__(key, value)

        if isinstance(value, Network):
            if hasattr(self, key) and isinstance(getattr(self, key), Network):
                raise SpaNetworkError(
                    "Cannot re-assign network-attribute %s to %s. SPA "
                    "network-attributes can only be assigned once."
                    % (key, value))

        super(Network, self).__setattr__(key, value)

        if isinstance(value, Network):
            self.__set_spa_network(key, value)

    def __set_spa_network(self, key, net):
        if net.label is None:
            net.label = key
        self._spa_networks[key] = net
        for k, (obj, v) in iteritems(net.inputs):
            if isinstance(v, int):
                net.inputs[k] = (obj, self.vocabs.get_or_create(v))
        for k, (obj, v) in iteritems(net.outputs):
            if isinstance(v, int):
                net.outputs[k] = (obj, self.vocabs.get_or_create(v))

    def get_spa_network(self, name, strip_output=False):
        """Return the SPA network for the given name.

        Raises :class:`SpaNetworkError` if the network cannot be found.

        Parameters
        ----------
        name : str
            Name of the network to retrieve.
        strip_output : bool, optional
            If ``True``, the network name is allowed to be followed by the name
            of an input or output that will be stripped (so the network with
            that input or output will be returned).

        Returns
        -------
        :class:`Network`
            Requested network.
        """
        try:
            components = name.split('.', 1)
            if len(components) > 1:
                head, tail = components
                return self._spa_networks[head].get_spa_network(
                    tail, strip_output=strip_output)
            else:
                if name in self._spa_networks:
                    return self._spa_networks[name]
                elif strip_output and (
                        name in self.inputs or name in self.outputs):
                    return self
                else:
                    raise KeyError
        except KeyError:
            raise SpaNetworkError("Could not find network %r." % name)

    def get_network_input(self, name):
        """Return the object to connect into for the given name.

        The name will be either the same as a spa network, or of the form
        <network_name>.<input_name>.
        """
        try:
            components = name.split('.', 1)
            if len(components) > 1:
                head, tail = components
                return self._spa_networks[head].get_network_input(tail)
            else:
                if name in self.inputs:
                    return self.inputs[name]
                elif name in self._spa_networks:
                    return self._spa_networks[name].get_network_input(
                        'default')
                else:
                    raise KeyError
        except KeyError:
            raise SpaNetworkError("Could not find network input %r." % name)

    def get_input_vocab(self, name):
        return self.get_network_input(name)[1]

    def get_network_output(self, name):
        """Return the object to connect into for the given name.

        The name will be either the same as a spa network, or of the form
        <network_name>.<output_name>.
        """
        try:
            components = name.split('.', 1)
            if len(components) > 1:
                head, tail = components
                return self._spa_networks[head].get_network_output(tail)
            else:
                if name in self.outputs:
                    return self.outputs[name]
                elif name in self._spa_networks:
                    return self._spa_networks[name].get_network_output(
                        'default')
                else:
                    raise KeyError
        except KeyError:
            raise SpaNetworkError("Could not find network output %r." % name)

    def get_output_vocab(self, name):
        return self.get_network_output(name)[1]

    def similarity(self, data, probe, vocab=None):
        """Return the similarity between the probed data and corresponding
        vocabulary.

        Parameters
        ----------
        data: ProbeDict
            Collection of simulation data returned by sim.run() function call.
        probe: Probe
            Probe with desired data.
        """
        if vocab is None:
            vocab = self.vocabs[data[probe].shape[1]]
        return nengo_spa.similarity(data[probe], vocab)
